from bson import ObjectId

from monkey_island.cc.database import mongo
import monkey_island.cc.services.node
from monkey_island.cc.models.monkey import Monkey, MonkeyNotFoundError

__author__ = "itay.mizeretz"


class EdgeService:
    def __init__(self):
        pass

    @staticmethod
    def get_displayed_edge_by_id(edge_id, for_report=False):
        edge = mongo.db.edge.find({"_id": ObjectId(edge_id)})[0]
        return EdgeService.edge_to_displayed_edge(edge, for_report)

    @staticmethod
    def get_displayed_edges_by_to(to, for_report=False):
        edges = mongo.db.edge.find({"to": ObjectId(to)})
        return [EdgeService.edge_to_displayed_edge(edge, for_report) for edge in edges]

    @staticmethod
    def edge_to_displayed_edge(edge, for_report=False):
        services = []
        os = {}

        if len(edge["scans"]) > 0:
            services = EdgeService.services_to_displayed_services(edge["scans"][-1]["data"]["services"], for_report)
            os = edge["scans"][-1]["data"]["os"]

        displayed_edge = EdgeService.edge_to_net_edge(edge)

        displayed_edge["ip_address"] = edge["ip_address"]
        displayed_edge["services"] = services
        displayed_edge["os"] = os
        displayed_edge["exploits"] = edge['exploits']
        displayed_edge["_label"] = EdgeService.get_edge_label(displayed_edge)
        return displayed_edge

    @staticmethod
    def insert_edge(from_id, to_id):
        edge_insert_result = mongo.db.edge.insert_one(
            {
                "from": from_id,
                "to": to_id,
                "scans": [],
                "exploits": [],
                "tunnel": False,
                "exploited": False,
                "src_label": EdgeService.get_label_for_endpoint(from_id),
                "dst_label": EdgeService.get_label_for_endpoint(to_id)
            })
        return mongo.db.edge.find_one({"_id": edge_insert_result.inserted_id})

    @staticmethod
    def get_or_create_edge(edge_from, edge_to):
        tunnel_edge = mongo.db.edge.find_one({"from": edge_from, "to": edge_to})
        if tunnel_edge is None:
            tunnel_edge = EdgeService.insert_edge(edge_from, edge_to)

        return tunnel_edge

    @staticmethod
    def generate_pseudo_edge(edge_id, edge_from, edge_to):
        edge = \
            {
                "id": edge_id,
                "from": edge_from,
                "to": edge_to,
                "group": "island",
                "src_label": EdgeService.get_label_for_endpoint(edge_from),
                "dst_label": EdgeService.get_label_for_endpoint(edge_to)
            }
        edge["_label"] = EdgeService.get_edge_label(edge)
        return edge

    @staticmethod
    def get_monkey_island_pseudo_edges():
        edges = []
        monkey_ids = [x["_id"] for x in mongo.db.monkey.find({}) if "tunnel" not in x]
        # We're using fake ids because the frontend graph module requires unique ids.
        # Collision with real id is improbable.
        count = 0
        for monkey_id in monkey_ids:
            count += 1
            edges.append(EdgeService.generate_pseudo_edge(
                ObjectId(hex(count)[2:].zfill(24)), monkey_id, ObjectId("000000000000000000000000")))

        return edges

    @staticmethod
    def get_infected_monkey_island_pseudo_edges():
        monkey = monkey_island.cc.services.node.NodeService.get_monkey_island_monkey()
        existing_ids = [x["from"] for x in mongo.db.edge.find({"to": monkey["_id"]})]
        monkey_ids = [x["_id"] for x in mongo.db.monkey.find({})
                      if ("tunnel" not in x) and (x["_id"] not in existing_ids) and (x["_id"] != monkey["_id"])]
        edges = []

        # We're using fake ids because the frontend graph module requires unique ids.
        # Collision with real id is improbable.
        count = 0
        for monkey_id in monkey_ids:
            count += 1
            edges.append(EdgeService.generate_pseudo_edge(
                ObjectId(hex(count)[2:].zfill(24)), monkey_id, monkey["_id"]))

        return edges

    @staticmethod
    def services_to_displayed_services(services, for_report=False):
        if for_report:
            return [x for x in services]
        else:
            return [x + ": " + (services[x]['name'] if 'name' in services[x] else 'unknown') for x in services]

    @staticmethod
    def edge_to_net_edge(edge):
        return \
            {
                "id": edge["_id"],
                "from": edge["from"],
                "to": edge["to"],
                "group": EdgeService.get_edge_group(edge),
                "src_label": edge["src_label"],
                "dst_label": edge["dst_label"]
            }

    @staticmethod
    def get_edge_group(edge):
        if edge.get("exploited"):
            return "exploited"
        if edge.get("tunnel"):
            return "tunnel"
        if (len(edge.get("scans", [])) > 0) or (len(edge.get("exploits", [])) > 0):
            return "scan"
        return "empty"

    @staticmethod
    def set_edge_exploited(edge):
        mongo.db.edge.update(
            {"_id": edge["_id"]},
            {"$set": {"exploited": True}}
        )
        monkey_island.cc.services.node.NodeService.set_node_exploited(edge["to"])

    @staticmethod
    def get_edge_label(edge):
        return "%s %s %s" % (edge['src_label'], RIGHT_ARROW, edge['dst_label'])

    @staticmethod
    def get_label_for_endpoint(endpoint_id):
        node_service = monkey_island.cc.services.node.NodeService
        if endpoint_id == ObjectId("000000000000000000000000"):
            return 'MonkeyIsland'
        if Monkey.is_monkey(endpoint_id):
            return Monkey.get_label_by_id(endpoint_id)
        else:
            return node_service.get_node_label(node_service.get_node_by_id(endpoint_id))

    @staticmethod
    def update_label_by_endpoint(edge, endpoint_id):
        label = EdgeService.get_label_for_endpoint(endpoint_id)
        if endpoint_id == edge["to"]:
            mongo_field = {"dst_label": label}
        else:
            mongo_field = {"src_label": label}
        mongo.db.edge.update({"_id": edge["_id"]},
                             {"$set": mongo_field})


RIGHT_ARROW = "\u2192"
